(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 12.2' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     16247,        435]
NotebookOptionsPosition[     15257,        408]
NotebookOutlinePosition[     15849,        429]
CellTagsIndexPosition[     15806,        426]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
    RowBox[{"Calculating", " ", "Bertrand"}], "-", 
    RowBox[{
    "Nash", " ", "and", " ", "Fully", " ", "Collusive", " ", "Prices"}]}], 
   "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"ClearAll", ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
      RowBox[{"Price", " ", "corresponding", " ", "to", " ", "i"}], "=", 
      "1"}], ",", "...", ",", 
     RowBox[{"15", "th", " ", "grid", " ", "point"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"p", "[", "i_", "]"}], ":=", 
     RowBox[{"1", "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"i", "-", "1"}], ")"}], 
       RowBox[{"(", 
        FractionBox["1.1", "14"], ")"}]}]}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "Profit", " ", "of", " ", "a", " ", "firm", " ", "charging", " ", 
     RowBox[{"p", "[", "i", "]"}], " ", "when", " ", "facing", " ", "one", 
     " ", "competitor", " ", "that", " ", "charges", " ", 
     RowBox[{"p", "[", "j", "]"}]}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"pi", "[", 
      RowBox[{"i_", ",", "j_", ",", "a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"p", "[", "i", "]"}], "-", "c"}], ")"}], 
      RowBox[{"(", 
       FractionBox[
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"a", "-", 
            RowBox[{"p", "[", "i", "]"}]}], ")"}], "/", "mu"}]], 
        RowBox[{"1", "+", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"a", "-", 
             RowBox[{"p", "[", "i", "]"}]}], ")"}], "/", "mu"}]], "+", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"a", "-", 
             RowBox[{"p", "[", "j", "]"}]}], ")"}], "/", "mu"}]]}]], 
       ")"}]}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{"Best", " ", "response", " ", "of", " ", "that", " ", "firm"}], 
     ",", " ", 
     RowBox[{"when", " ", "the", " ", "competitor", " ", "charges", " ", 
      RowBox[{"p", "[", "j", "]"}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"br", "[", 
      RowBox[{"j_", ",", "a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"i", "/.", 
      RowBox[{"Last", "[", 
       RowBox[{"NMaximize", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"pi", "[", 
            RowBox[{"i", ",", "j", ",", "a", ",", "mu", ",", "c"}], "]"}], 
           ",", 
           RowBox[{"i", "\[Element]", "Integers"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "1", ",", "15"}], "}"}]}], "]"}], "]"}]}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "Check", " ", "if", " ", "there", " ", "is", " ", "a", " ", "Nash", " ", 
     "equilibrium", " ", "where", " ", "one", " ", "firm", " ", "chooses", 
     " ", "the", " ", "jth", " ", "element", " ", "in", " ", "the", " ", 
     "pricing", " ", "support"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"NE", "[", 
      RowBox[{"j_", ",", "a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"TrueQ", "[", 
      RowBox[{
       RowBox[{"br", "[", 
        RowBox[{
         RowBox[{"br", "[", 
          RowBox[{"j", ",", "a", ",", "mu", ",", "c"}], "]"}], ",", "a", ",", 
         "mu", ",", "c"}], "]"}], "\[Equal]", "j"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{"Print", " ", "the", " ", "Bertrand"}], "-", 
     RowBox[{"Nash", " ", "equilibrium", " ", "prices"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"printNE", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"Do", "[", 
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"NE", "[", 
           RowBox[{"j", ",", "a", ",", "mu", ",", "c"}], "]"}], "==", 
          "True"}], ",", 
         RowBox[{"Print", "[", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"p", "[", "j", "]"}], ",", 
            RowBox[{"p", "[", 
             RowBox[{"br", "[", 
              RowBox[{"j", ",", "a", ",", "mu", ",", "c"}], "]"}], "]"}]}], 
           "}"}], "]"}], ",", 
         StyleBox["\"\<\>\"",
          FontWeight->"Bold"]}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"j", ",", "1", ",", "15"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"Fully", " ", "collusive", " ", "grid", " ", "points"}], " ", 
    "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"FC1", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"i", "/.", 
      RowBox[{"Last", "[", 
       RowBox[{"NMaximize", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            RowBox[{"pi", "[", 
             RowBox[{"i", ",", "j", ",", "a", ",", "mu", ",", "c"}], "]"}], 
            "+", 
            RowBox[{"pi", "[", 
             RowBox[{"j", ",", "i", ",", "a", ",", "mu", ",", "c"}], "]"}]}], 
           ",", 
           RowBox[{"i", "\[Element]", "Integers"}], ",", 
           RowBox[{"j", "\[Element]", "Integers"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"i", ",", "1", ",", "15"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"j", ",", "1", ",", "15"}], "}"}]}], "}"}]}], "]"}], 
       "]"}]}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"FC2", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"j", "/.", 
      RowBox[{"Last", "[", 
       RowBox[{"NMaximize", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            RowBox[{"pi", "[", 
             RowBox[{"i", ",", "j", ",", "a", ",", "mu", ",", "c"}], "]"}], 
            "+", 
            RowBox[{"pi", "[", 
             RowBox[{"j", ",", "i", ",", "a", ",", "mu", ",", "c"}], "]"}]}], 
           ",", 
           RowBox[{"i", "\[Element]", "Integers"}], ",", 
           RowBox[{"j", "\[Element]", "Integers"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"i", ",", "1", ",", "15"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"j", ",", "1", ",", "15"}], "}"}]}], "}"}]}], "]"}], 
       "]"}]}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"And", " ", "so", " ", "fully", " ", "collusive", " ", "prices"}],
     " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"pFC1", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"p", "[", 
      RowBox[{"FC1", "[", 
       RowBox[{"a", ",", "mu", ",", "c"}], "]"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"pFC2", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"p", "[", 
      RowBox[{"FC2", "[", 
       RowBox[{"a", ",", "mu", ",", "c"}], "]"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"Explicitly", " ", "calculate", " ", "collusive", " ", "profit"}],
     " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"checkFC", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"NMaximize", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"pi", "[", 
           RowBox[{"i", ",", "j", ",", "a", ",", "mu", ",", "c"}], "]"}], "+", 
          RowBox[{"pi", "[", 
           RowBox[{"j", ",", "i", ",", "a", ",", "mu", ",", "c"}], "]"}]}], 
         ",", 
         RowBox[{"i", "\[Element]", "Integers"}], ",", 
         RowBox[{"j", "\[Element]", "Integers"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"i", ",", "1", ",", "15"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"j", ",", "1", ",", "15"}], "}"}]}], "}"}]}], "]"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "Derive", " ", "the", " ", "best", " ", "profit", " ", "not", " ", 
     "obtained", " ", "at", " ", "the", " ", "solution", " ", "just", " ", 
     "found"}], "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"FC2ndbest", "[", 
      RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
     RowBox[{"NMaximize", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"pi", "[", 
           RowBox[{"i", ",", "j", ",", "a", ",", "mu", ",", "c"}], "]"}], "+", 
          RowBox[{"pi", "[", 
           RowBox[{"j", ",", "i", ",", "a", ",", "mu", ",", "c"}], "]"}]}], 
         ",", 
         RowBox[{"i", "\[Element]", "Integers"}], ",", 
         RowBox[{"j", "\[Element]", "Integers"}], ",", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"i", ",", "j"}], "}"}], "!=", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"FC1", "[", 
             RowBox[{"a", ",", "mu", ",", "c"}], "]"}], ",", 
            RowBox[{"FC2", "[", 
             RowBox[{"a", ",", "mu", ",", "c"}], "]"}]}], "}"}]}]}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"i", ",", "1", ",", "15"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"j", ",", "1", ",", "15"}], "}"}]}], "}"}]}], "]"}]}], 
    ";"}], "\[IndentingNewLine]", 
   StyleBox[
    RowBox[{"(*", " ", 
     RowBox[{
     "Then", " ", "check", " ", "what", " ", "we", " ", "found", " ", 
      "strictly", " ", "beats", " ", "this"}], " ", "*)"}],
    FontWeight->"Plain"], 
   StyleBox["\[IndentingNewLine]",
    FontWeight->"Plain"], 
   StyleBox[
    RowBox[{
     RowBox[{
      RowBox[{"uniqueFC", "[", 
       RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
      RowBox[{"TrueQ", "[", 
       RowBox[{
        RowBox[{"First", "[", 
         RowBox[{"checkFC", "[", 
          RowBox[{"a", ",", "mu", ",", "c"}], "]"}], "]"}], ">", 
        RowBox[{"First", "[", 
         RowBox[{"FC2ndbest", "[", 
          RowBox[{"a", ",", "mu", ",", "c"}], "]"}], "]"}]}], "]"}]}], ";"}],
    FontWeight->"Bold"], 
   StyleBox["\[IndentingNewLine]",
    FontWeight->"Bold"], 
   StyleBox[
    RowBox[{
     RowBox[{
      RowBox[{"printFC", "[", 
       RowBox[{"a_", ",", "mu_", ",", "c_"}], "]"}], ":=", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"uniqueFC", "[", 
          RowBox[{"a", ",", "mu", ",", "c"}], "]"}], "==", "True"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"pFC1", "[", 
           RowBox[{"a", ",", "mu", ",", "c"}], "]"}], ",", 
          RowBox[{"pFC2", "[", 
           RowBox[{"a", ",", "mu", ",", "c"}], "]"}]}], "}"}], ",", 
        "\"\<\>\""}], "]"}]}], ";"}],
    FontWeight->"Bold"], 
   StyleBox["\[IndentingNewLine]",
    FontWeight->"Bold"], 
   StyleBox[
    RowBox[{
    "Print", "[", 
     "\"\<(* Table 1, Column 1: compute Bertrand-Nash Prices *)\>\"", "]"}],
    FontWeight->"Bold"], 
   StyleBox["\[IndentingNewLine]",
    FontWeight->"Bold"], 
   RowBox[{"printNE", "[", 
    RowBox[{"2", ",", "0.25", ",", "1"}], "]"}], "\[IndentingNewLine]", 
   RowBox[{
   "Print", "[", "\"\<(* Table 1, Column 2: compute Collusive Prices *)\>\"", 
    "]"}], "\[IndentingNewLine]", 
   RowBox[{"printFC", "[", 
    RowBox[{"2", ",", "0.25", ",", "1"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.868426478788829*^9, 3.868426523341015*^9}, {
   3.8684266056786423`*^9, 3.868426624707316*^9}, {3.868426663697239*^9, 
   3.868426730672679*^9}, {3.8684274447053547`*^9, 3.868427473653738*^9}, {
   3.8684276175248947`*^9, 3.868427653313505*^9}, {3.868428000018187*^9, 
   3.868428033422567*^9}, {3.8684280703503304`*^9, 3.868428113581561*^9}, {
   3.8722395185460296`*^9, 3.8722395204332185`*^9}, {3.8736431799394026`*^9, 
   3.8736432106677003`*^9}, {3.8736437278566628`*^9, 
   3.8736437677845945`*^9}, {3.873643802209982*^9, 3.873643869297222*^9}, {
   3.8736439209031553`*^9, 3.873644020926572*^9}, {3.8736444150433264`*^9, 
   3.8736445288646784`*^9}, 3.8736445782855306`*^9, {3.8736446297392206`*^9, 
   3.873644680110836*^9}, {3.873644713328519*^9, 3.8736447769333367`*^9}, {
   3.8736448385702295`*^9, 3.8736448586463833`*^9}, {3.87364489399482*^9, 
   3.8736449171887164`*^9}, {3.8736449516911983`*^9, 
   3.8736449843834305`*^9}, {3.87364563769258*^9, 3.873645646672364*^9}, {
   3.8853748808892183`*^9, 3.8853748953838944`*^9}, {3.8853752607627134`*^9, 
   3.8853752616885967`*^9}, {3.885376442405448*^9, 3.8853764499726477`*^9}, {
   3.8853787003046923`*^9, 3.8853788687200155`*^9}, {3.885378922639926*^9, 
   3.8853789326963854`*^9}, {3.8924761438281937`*^9, 3.892476156148532*^9}, {
   3.892476193553672*^9, 3.892476236186906*^9}, {3.8924766244996758`*^9, 
   3.8924766607287617`*^9}, {3.8924767269906178`*^9, 3.8924767962188845`*^9}, 
   3.8924768528445206`*^9, {3.8924769070689936`*^9, 3.8924769512107277`*^9}, {
   3.892699531243026*^9, 3.892699632225586*^9}, {3.8926999820721397`*^9, 
   3.892699988707426*^9}, {3.8927084435482855`*^9, 3.8927085324217234`*^9}, 
   3.8927085645916185`*^9, {3.892708634012557*^9, 3.8927086943721557`*^9}},
 CellLabel->"In[72]:=",ExpressionUUID->"1e8b95c2-23d3-49fa-a7b2-aea59f2a2829"],

Cell[CellGroupData[{

Cell[BoxData["\<\"(* Table 1, Column 1: compute Bertrand-Nash Prices \
*)\"\>"], "Print",
 CellChangeTimes->{3.892708695709658*^9, 3.892708872365257*^9},
 CellLabel->
  "During evaluation of \
In[72]:=",ExpressionUUID->"6ffd2ecc-f733-4728-a5e6-5dfcb1baed66"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.4714285714285715`", ",", "1.4714285714285715`"}], "}"}]], "Print",\

 CellChangeTimes->{3.892708695709658*^9, 3.8927088729984565`*^9},
 CellLabel->
  "During evaluation of \
In[72]:=",ExpressionUUID->"a5b6687d-c290-4509-991d-a0e9332fb23d"],

Cell[BoxData["\<\"(* Table 1, Column 2: compute Collusive Prices *)\"\>"], \
"Print",
 CellChangeTimes->{3.892708695709658*^9, 3.8927088737091303`*^9},
 CellLabel->
  "During evaluation of \
In[72]:=",ExpressionUUID->"bfe5fff7-27cc-4f66-842f-3e3563078709"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.9428571428571428`", ",", "1.9428571428571428`"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.892708697846067*^9, 3.892708874490796*^9},
 CellLabel->"Out[89]=",ExpressionUUID->"dd503fba-6782-466a-9daf-eb1319f21c94"]
}, Open  ]]
},
WindowSize->{1428., 729.75},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"Magnification"->1.,
"PaperOrientation"->"Portrait",
"PaperSize"->{595.2755905511812, 841.8897637795276}},
FrontEndVersion->"13.2 for Microsoft Windows (64-bit) (November 18, 2022)",
StyleDefinitions->"Default.nb",
ExpressionUUID->"21e771e4-d3c4-4778-8571-417fca78ae06"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 13565, 350, 609, "Input",ExpressionUUID->"1e8b95c2-23d3-49fa-a7b2-aea59f2a2829"],
Cell[CellGroupData[{
Cell[14170, 376, 258, 5, 22, "Print",ExpressionUUID->"6ffd2ecc-f733-4728-a5e6-5dfcb1baed66"],
Cell[14431, 383, 281, 7, 22, "Print",ExpressionUUID->"a5b6687d-c290-4509-991d-a0e9332fb23d"],
Cell[14715, 392, 256, 5, 22, "Print",ExpressionUUID->"bfe5fff7-27cc-4f66-842f-3e3563078709"]
}, Open  ]],
Cell[14986, 400, 255, 5, 32, "Output",ExpressionUUID->"dd503fba-6782-466a-9daf-eb1319f21c94"]
}, Open  ]]
}
]
*)

